/*
 *  Linpack-in-C: Software for Numerical Linear Algebra.
 *  Copyright (C) 1996 Ross Ihaka
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Linpack.h>

/*
 *     DPTSL given a positive definite tridiagonal matrix and a right
 *     hand side will find the solution.
 *
 *     On Entry
 *
 *        n        int
 *                 is the order of the tridiagonal matrix.
 *
 *        d        double(n)
 *                 is the diagonal of the tridiagonal matrix.
 *                 on output d is destroyed.
 *
 *        e        double(n)
 *                 is the offdiagonal of the tridiagonal matrix.
 *                 e(1) through e(n-1) should contain the
 *                 offdiagonal.
 *
 *        b        double(n)
 *                 is the right hand side vector.
 *
 *     On Return
 *
 *        b        contains the soultion.
 *
 *     LINPACK. This version dated 08/14/78.
 *     Jack Dongarra, Argonne National Laboratory.
 *     C Translation by Ross Ihaka.
 */

#define d(i)	d[i]
#define e(i)	e[i]
#define b(i)	b[i]

void DPTSL(int n, double *d, double *e, double *b)
{
	int k, kbm1, ke, kf, kp1, nm1, nm1d2;
	double t1, t2;

	d -= 1;
	e -= 1;
	b -= 1;

	/* check for 1 x 1 case */

	if(n == 1)
		b(1) = b(1)/d(1);
	else {
		nm1 = n-1;
		nm1d2 = nm1/2;
		if(n != 2) {
			kbm1 = n-1;

			/* zero top half of subdiagonal and bottom half of */
			/* superdiagonal */

			for(k=1 ; k <= nm1d2  ; k++) {
				t1 = e(k)/d(k);
				d(k+1) = d(k+1)-t1*e(k);
				b(k+1) = b(k+1)-t1*b(k);
				t2 = e(kbm1)/d(kbm1+1);
				d(kbm1) = d(kbm1)-t2*e(kbm1);
				b(kbm1) = b(kbm1)-t2*b(kbm1+1);
				kbm1 = kbm1-1;
			}
		}
		kp1 = nm1d2+1;

		/* clean up for possible 2 x 2 block at center */

		if((n % 2) == 0) {
			t1 = e(kp1)/d(kp1);
			d(kp1+1) = d(kp1+1)-t1*e(kp1);
			b(kp1+1) = b(kp1+1)-t1*b(kp1);
			kp1 = kp1+1;
		}

		/* back solve starting at the center, going towards the top */
		/* and bottom */

		b(kp1) = b(kp1)/d(kp1);
		if(n != 2) {
			k = kp1-1;
			ke = kp1+nm1d2-1;
			for(kf=kp1 ; kf <= ke  ; kf++) {
				b(k) = (b(k)-e(k)*b(k+1))/d(k);
				b(kf+1) = (b(kf+1)-e(kf)*b(kf))/d(kf+1);
				k = k-1;
			}
		}
		if((n % 2) == 0)
			b(1) = (b(1)-e(1)*b(2))/d(1);
	}
	return;
}

int dptsl_(int *n, double *d, double *e, double *b)
{
	DPTSL(*n, d, e, b);
}
