/*
 *  Linpack-in-C: Software for Numerical Linear Algebra.
 *  Copyright (C) 1996 Ross Ihaka
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Linpack.h>

/*
 *     DSVDC is a subroutine to reduce a double nxp matrix x
 *     by orthogonal transformations u and v to diagonal form.  the
 *     diagonal elements s(i) are the singular values of x.  the
 *     columns of u are the corresponding left singular vectors,
 *     and the columns of v the right singular vectors.
 *
 *     On Entry
 *
 *         x         double(ldx,p), where ldx>=n.
 *                   x contains the matrix whose singular value
 *                   decomposition is to be computed.  x is
 *                   destroyed by dsvdc.
 *
 *         ldx       int.
 *                   ldx is the leading dimension of the array x.
 *
 *         n         int.
 *                   n is the number of rows of the matrix x.
 *
 *         p         int.
 *                   p is the number of columns of the matrix x.
 *
 *         ldu       int.
 *                   ldu is the leading dimension of the array u.
 *                   (see below).
 *
 *         ldv       int.
 *                   ldv is the leading dimension of the array v.
 *                   (see below).
 *
 *         work      double(n).
 *                   work is a scratch array.
 *
 *         job       int.
 *                   job controls the computation of the singular
 *                   vectors.  it has the decimal expansion ab
 *                   with the following meaning
 *
 *                        a==0    do not compute the left singular
 *                                  vectors.
 *                        a==1    return the n left singular vectors
 *                                  in u.
 *                        a>=2    return the first min(n,p) singular
 *                                  vectors in u.
 *                        b==0    do not compute the right singular
 *                                  vectors.
 *                        b==1    return the right singular vectors
 *                                  in v.
 *
 *     On Return
 *
 *         s         double(mm), where mm=min(n+1,p).
 *                   the first min(n,p) entries of s contain the
 *                   singular values of x arranged in descending
 *                   order of magnitude.
 *
 *         e         double(p),
 *                   e ordinarily contains zeros.  however see the
 *                   discussion of info for exceptions.
 *
 *         u         double(ldu,k), where ldu>=n.  if
 *                                   joba==1 then k==n, if joba>=2
 *                                   then k==min(n,p).
 *                   u contains the matrix of left singular vectors.
 *                   u is not referenced if joba==0.  if n<=p
 *                   or if joba==2, then u may be identified with x
 *                   in the subroutine call.
 *
 *         v         double(ldv,p), where ldv>=p.
 *                   v contains the matrix of right singular vectors.
 *                   v is not referenced if job==0.  if p<=n,
 *                   then v may be identified with x in the
 *                   subroutine call.
 *
 *         info      int.
 *                   the singular values (and their corresponding
 *                   singular vectors) s(info+1),s(info+2),...,s(m)
 *                   are correct (here m=min(n,p)).  thus if
 *                   info==0, all the singular values and their
 *                   vectors are correct.  in any event, the matrix
 *                   b = trans(u)*x*v is the bidiagonal matrix
 *                   with the elements of s on its diagonal and the
 *                   elements of e on its super-diagonal (trans(u)
 *                   is the transpose of u).  thus the singular
 *                   values of x and b are the same.
 *
 *     LINPACK. This version dated 08/14/78.
 *              Correction made to shift 2/84.
 *     G.W. Stewart, University of Maryland, Argonne National Lab.
 *     C Translation by Ross Ihaka.
 */

#define x(i,j)	x[i+(j)*ldx]
#define u(i,j)	u[i+(j)*ldu]
#define v(i,j)	v[i+(j)*ldv]
#define s(i)	s[i]
#define e(i)	e[i]
#define work(i)	work[i]

#define TRUE	1
#define FALSE	0

void DSVDC(double *x, int ldx, int n, int p, double *s, double *e, double *u, int ldu, double *v, int ldv, double *work, int job, int *info)
{
	int i, iter, j, jobu, k, kase, kk, l, ll, lls, lm1, lp1, ls, lu, m, maxit, mm, mm1, mp1, nct, nctp1, ncu, nrt, nrtp1;
	double t, r;
	double b, c, cs, el, emm1, f, g, scale, shift, sl, sm, sn, smm1, t1, test, ztest;
	int wantu, wantv;

	x -= (ldx+1);
	u -= (ldu+1);
	v -= (ldv+1);
	s -= 1;
	e -= 1;
	work -= 1;

	/* set the maximum number of iterations. */

	maxit = 30;

	/* determine what is to be computed. */

	wantu = FALSE;
	wantv = FALSE;
	jobu = (job % 100)/10;
	ncu = n;
	if(jobu > 1)
		ncu = imin(n, p);
	if(jobu != 0)
		wantu = TRUE;
	if((job % 10) != 0)
		wantv = TRUE;

	/* reduce x to bidiagonal form, storing the diagonal elements */
	/* in s and the super-diagonal elements in e. */

	*info = 0;
	nct = imin(n-1, p);
	nrt = imax(0, imin(p-2, n));
	lu = imax(nct, nrt);
	if(lu >= 1)
		for(l=1 ; l <= lu  ; l++) {
			lp1 = l+1;
			if(l <= nct) {

				/* compute the transformation for the l-th column and */
				/* place the l-th diagonal in s(l). */

				s(l) = DNRM2(n-l+1, &x(l, l), 1);
				if(s(l) != 0.0) {
					if(x(l, l) != 0.0)
						s(l) = fsign(s(l), x(l, l));
					DSCAL(n-l+1, 1.0/s(l), &x(l, l), 1);
					x(l, l) = 1.0+x(l, l);
				}
				s(l) = -s(l);
			}
			if(p >= lp1)
				for(j=lp1 ; j <= p  ; j++) {
					if(l <= nct)
						if(s(l) != 0.0) {

							/* apply the transformation. */

							t = -DDOT(n-l+1, &x(l, l), 1, &x(l, j), 1)/x(l, l);
							DAXPY(n-l+1, t, &x(l, l), 1, &x(l, j), 1);
						}

						/* place the l-th row of x into  e for the */
						/* subsequent calculation of the row transformation. */

					e(j) = x(l, j);
				}
			if(wantu && l <= nct)

				/* place the transformation in u for subsequent back */
				/* multiplication. */

				for(i=l ; i <= n ; i++) 
					u(i, l) = x(i, l);
			if(l <= nrt) {

				/* compute the l-th row transformation and place the */
				/* l-th super-diagonal in e(l). */

				e(l) = DNRM2(p-l, &e(lp1), 1);
				if(e(l) != 0.0) {
					if(e(lp1) != 0.0)
						e(l) = fsign(e(l), e(lp1));
					DSCAL(p-l, 1.0/e(l), &e(lp1), 1);
					e(lp1) = 1.0+e(lp1);
				}
				e(l) = -e(l);
				if(lp1 <= n && e(l) != 0.0) {

					/* apply the transformation. */

					for(i=lp1 ; i <= n ; i++) 
						work(i) = 0.0;
					for(j=lp1 ; j <= p ; j++) 
						DAXPY(n-l, e(j), &x(lp1, j), 1, &work(lp1), 1);
					for(j=lp1 ; j <= p ; j++) 
						DAXPY(n-l, -e(j)/e(lp1), &work(lp1), 1, &x(lp1, j), 1);
				}
				if(wantv)

					/* place the transformation in v for subsequent */
					/* back multiplication. */

					for(i=lp1 ; i <= p ; i++) 
						v(i, l) = e(i);
			}
		}

	/* set up the final bidiagonal matrix or order m. */

	m = imin(p, n+1);
	nctp1 = nct+1;
	nrtp1 = nrt+1;
	if(nct < p)
		s(nctp1) = x(nctp1, nctp1);
	if(n < m)
		s(m) = 0.0;
	if(nrtp1 < m)
		e(nrtp1) = x(nrtp1, m);
	e(m) = 0.0;

	/* if required, generate u. */

	if(wantu) {
		if(ncu >= nctp1)
			for(j=nctp1 ; j <= ncu  ; j++) {
				for(i=1 ; i <= n ; i++) 
					u(i, j) = 0.0;
				u(j, j) = 1.0;
			}
		if(nct >= 1)
			for(ll=1 ; ll <= nct  ; ll++) {
				l = nct-ll+1;
				if(s(l) == 0.0) {
					for(i=1 ; i <= n ; i++) 
						u(i, l) = 0.0;
					u(l, l) = 1.0;
				}
				else {
					lp1 = l+1;
					if(ncu >= lp1)
						for(j=lp1 ; j <= ncu  ; j++) {
							t = -DDOT(n-l+1, &u(l, l), 1, &u(l, j), 1)/u(l, l);
							DAXPY(n-l+1, t, &u(l, l), 1, &u(l, j), 1);
						}
					DSCAL(n-l+1, -1.0, &u(l, l), 1);
					u(l, l) = 1.0+u(l, l);
					lm1 = l-1;
					if(lm1 >= 1)
						for(i=1 ; i <= lm1 ; i++) 
							u(i, l) = 0.0;
				}
			}
	}

	/* if it is required, generate v. */

	if(wantv)
		for(ll=1 ; ll <= p  ; ll++) {
			l = p-ll+1;
			lp1 = l+1;
			if(l <= nrt)
				if(e(l) != 0.0)
					for(j=lp1 ; j <= p  ; j++) {
						t = -DDOT(p-l, &v(lp1, l), 1, &v(lp1, j), 1)/v(lp1, l);
						DAXPY(p-l, t, &v(lp1, l), 1, &v(lp1, j), 1);
					}
			for(i=1 ; i <= p ; i++) 
				v(i, l) = 0.0;
			v(l, l) = 1.0;
		}

	/* main iteration loop for the singular values. */

	mm = m;
	iter = 0;
	for(;;) {

		/* quit if all the singular values have been found. */

		if(m == 0)
			return;

		/* if too many iterations have been performed, set */
		/* flag and return. */

		if(iter >= maxit)
			break;

		/* this section of the program inspects for */
		/* negligible elements in the s and e arrays.  on */
		/* completion the variables kase and l are set as follows. */

		/* kase = 1     if s(m) and e(l-1) are negligible and l<m */
		/* kase = 2     if s(l) is negligible and l<m */
		/* kase = 3     if e(l-1) is negligible, l<m, and */
		/* s(l),..., s(m) are not negligible (qr step). */
		/* kase = 4     if e(m-1) is negligible (convergence). */

		for(ll=1 ; ll <= m  ; ll++) {
			l = m-ll;
			if(l == 0)
				break;
			test = fabs(s(l))+fabs(s(l+1));
			ztest = test+fabs(e(l));
			if(ztest == test)
				goto L10;
		}
		goto L20;
	L10:	e(l) = 0.0;
	L20:	if(l == m-1)
			kase = 4;
		else {
			lp1 = l+1;
			mp1 = m+1;
			for(lls=lp1 ; lls <= mp1  ; lls++) {
				ls = m-lls+lp1;
				if(ls == l)
					break;
				test = 0.0;
				if(ls != m)
					test = test+fabs(e(ls));
				if(ls != l+1)
					test = test+fabs(e(ls-1));
				ztest = test+fabs(s(ls));
				if(ztest == test)
					goto L30;
			}
			goto L40;
		L30:	s(ls) = 0.0;
		L40:	if(ls == l)
				kase = 3;
			else if(ls == m)
				kase = 1;
			else {
				kase = 2;
				l = ls;
			}
		}
		l = l+1;

		/* perform the task indicated by kase. */

		switch(kase) {
			case 1:
				mm1 = m-1;
				f = e(m-1);
				e(m-1) = 0.0;
				for(kk=l ; kk <= mm1  ; kk++) {
					k = mm1-kk+l;
					t1 = s(k);
					DROTG(&t1, &f, &cs, &sn);
					s(k) = t1;
					if(k != l) {
						f = -sn*e(k-1);
						e(k-1) = cs*e(k-1);
					}
					if(wantv)
						DROT(p, &v(1, k), 1, &v(1, m), 1, cs, sn);
				}
				break;
			case 2:
				f = e(l-1);
				e(l-1) = 0.0;
				for(k=l ; k <= m  ; k++) {
					t1 = s(k);
					DROTG(&t1, &f, &cs, &sn);
					s(k) = t1;
					f = -sn*e(k);
					e(k) = cs*e(k);
					if(wantu)
						DROT(n, &u(1, k), 1, &u(1, l-1), 1, cs, sn);
				}
				break;
			case 3:

				/* calculate the shift. */

				scale = fmax(fabs(s(m)), fmax(fabs(s(m-1)), fmax(fabs(e(m-1)), fmax(fabs(s(l)), fabs(e(l))))));
				sm = s(m)/scale;
				smm1 = s(m-1)/scale;
				emm1 = e(m-1)/scale;
				sl = s(l)/scale;
				el = e(l)/scale;
				b = ((smm1+sm)*(smm1-sm)+emm1*emm1)/2.0;
				c = fsquare(sm*emm1);
				shift = 0.0;
				if(b != 0.0 || c != 0.0) {
					shift = sqrt(b*b+c);
					if(b < 0.0)
						shift = -shift;
					shift = c/(b+shift);
				}
				f = (sl+sm)*(sl-sm)+shift;
				g = sl*el;

				/* chase zeros. */

				mm1 = m-1;
				for(k=l ; k <= mm1  ; k++) {
					DROTG(&f, &g, &cs, &sn);
					if(k != l)
						e(k-1) = f;
					f = cs*s(k)+sn*e(k);
					e(k) = cs*e(k)-sn*s(k);
					g = sn*s(k+1);
					s(k+1) = cs*s(k+1);
					if(wantv)
						DROT(p, &v(1, k), 1, &v(1, k+1), 1, cs, sn);
					DROTG(&f, &g, &cs, &sn);
					s(k) = f;
					f = cs*e(k)+sn*s(k+1);
					s(k+1) = -sn*e(k)+cs*s(k+1);
					g = sn*e(k+1);
					e(k+1) = cs*e(k+1);
					if(wantu && k < n)
						DROT(n, &u(1, k), 1, &u(1, k+1), 1, cs, sn);
				}
				e(m-1) = f;
				iter = iter+1;
				break;
			case 4:

				/* make the singular value  positive. */

				if(s(l) < 0.0) {
					s(l) = -s(l);
					if(wantv)
						DSCAL(p, -1.0, &v(1, l), 1);
				}
				while (l != mm) {
					if(s(l) >= s(l+1))
						break;
					t = s(l);
					s(l) = s(l+1);
					s(l+1) = t;
					if(wantv && l < p)
						DSWAP(p, &v(1, l), 1, &v(1, l+1), 1);
					if(wantu && l < n)
						DSWAP(n, &u(1, l), 1, &u(1, l+1), 1);
					l = l+1;
				}
				iter = 0;
				m = m-1;
		}
	}
	*info = m;
	return;
}

int dsvdc_(double *x, int *ldx, int *n, int *p, double *s, double *e, double *u, int *ldu, double *v, int *ldv, double *work, int *job, int *info)
{
	DSVDC(x, *ldx, *n, *p, s, e, u, *ldu, v, *ldv, work, *job, info);
}
